import click
from click.decorators import pass_context
from flask import current_app
from flask.cli import with_appcontext
from importlib_metadata import EntryPoint, entry_points

from flaschengeist.database import db
from flaschengeist.config import config
from flaschengeist.models.user import Permission


@click.group()
def plugin():
    pass


def install_plugin_command(ctx, plugin, all):
    """Install one or more plugins"""
    if not all and len(plugin) == 0:
        ctx.fail("At least one plugin must be specified, or use `--all` flag.")

    if all:
        plugins = current_app.config["FG_PLUGINS"]
    else:
        try:
            plugins = {plugin_name: current_app.config["FG_PLUGINS"][plugin_name] for plugin_name in plugin}
        except KeyError as e:
            ctx.fail(f"Invalid plugin name, could not find >{e.args[0]}<")

    for name, plugin in plugins.items():
        click.echo(f"Installing {name}{'.'*(20-len(name))}", nl=False)
        # Install permissions
        if plugin.permissions:
            cur_perm = set(x.name for x in Permission.query.filter(Permission.name.in_(plugin.permissions)).all())
            all_perm = set(plugin.permissions)

            add = all_perm - cur_perm
            if add:
                db.session.bulk_save_objects([Permission(name=x) for x in all_perm])
                db.session.commit()
        # Custom installation steps
        plugin.install()
        click.secho(" ok", fg="green")


@plugin.command()
@click.argument("plugin", nargs=-1, type=str)
@click.option("--all", help="Install all enabled plugins", is_flag=True)
@with_appcontext
@pass_context
def install(ctx, plugin, all):
    """Install one or more plugins"""
    return install_plugin_command(ctx, plugin, all)


@plugin.command()
@click.argument("plugin", nargs=-1, type=str)
@with_appcontext
@pass_context
def uninstall(ctx: click.Context, plugin):
    """Uninstall one or more plugins"""

    if len(plugin) == 0:
        ctx.fail("At least one plugin must be specified")

    try:
        plugins = {plugin_name: current_app.config["FG_PLUGINS"][plugin_name] for plugin_name in plugin}
    except KeyError as e:
        ctx.fail(f"Invalid plugin ID, could not find >{e.args[0]}<")

    if (
        click.prompt(
            "You are going to uninstall:\n\n"
            f"\t{', '.join([plugin_name for plugin_name in plugins.keys()])}\n\n"
            "Are you sure?",
            default="n",
            show_choices=True,
            type=click.Choice(["y", "N"], False),
        ).lower()
        != "y"
    ):
        ctx.exit()
    for name, plugin in plugins.items():
        click.echo(f"Uninstalling {name}{'.'*(20-len(name))}", nl=False)
        plugin.uninstall()
        click.secho(" ok", fg="green")


@plugin.command()
@click.option("--enabled", "-e", help="List only enabled plugins", is_flag=True)
@click.option("--no-header", "-n", help="Do not show header", is_flag=True)
@with_appcontext
def ls(enabled, no_header):
    def plugin_version(p):
        if isinstance(p, EntryPoint):
            return p.dist.version
        return p.version

    plugins = entry_points(group="flaschengeist.plugins")
    enabled_plugins = [key for key, value in config.items() if "enabled" in value and value["enabled"]] + [
        config["FLASCHENGEIST"]["auth"]
    ]
    loaded_plugins = current_app.config["FG_PLUGINS"].keys()

    if not no_header:
        print(f"{' '*13}{'name': <20}|{'version': >10}")
        print("-" * 46)
    for plugin in plugins:
        if enabled and plugin.name not in enabled_plugins:
            continue
        print(
            f"{plugin.name: <33}|{plugin_version(plugin): >12}"
            f"{click.style(' (enabled)', fg='green') if plugin.name in enabled_plugins else ' (disabled)'}"
        )

    for plugin in [value for value in enabled_plugins if value not in loaded_plugins]:
        print(f"{plugin: <33}|{' '*12}" f"{click.style(' (not found)', fg='red')}")
